/* FILE: loop.c                 Copyright (c), Symmetric Research, 1986 1998
 *
 * Looped radix 2, decimation in time, in place FFT.
 *
 *     F[k] = sum( A[n] * cexp( isign*2*PI*n*k/N))    n,k = 0,1,...,N-1
 *
 *
 *
 * NOTES:
 *
 * This code is a straightforward theoretical implementation of a
 * non recursive radix 2 FFT algorithm.  Note that it has many
 * redundant trig evaluations in its innermost butterfly loop.
 *
 * More typical production code is listed in r2dit.c.  The code given
 * there avoids redundant trig evaluations, and uses small tables to
 * remove the need for calling any trig functions at all.
 *
 * See the code fragments listed at the bottom of this file for the
 * steps required to progress from the code given here to r2dit.c.
 *
 *
 *
 *
 */

#include <math.h>
#include "fft.h"



void FFT( COMPLEX *A, int N, int isign) {

        int n, i, j, mask; COMPLEX temp, *A1, *A2, cexp;
        


        /* Scramble the input data into bit reversed order. */

        for ( i = 0 , j = 0 ; i < N ; i++ ) {


                /* Swap A[i] and A[j] if not done yet. */

                if ( j > i ) {

                        A1 = &A[i];
                        A2 = &A[j];
                        
                        temp.r = A1->r;
                        temp.i = A1->i;
                        
                        A1->r = A2->r;
                        A1->i = A2->i;
                        
                        A2->r = temp.r;
                        A2->i = temp.i;

                        }

                /* Reverse inc j. */

                for ( mask = N/2 ; mask&j ; mask >>= 1 )

                        j &= ~mask;

                j |= mask;

                }




        /* Compute the FFT butterflies. */

        for ( n = 2 ; n <= N ; n *= 2 ) {

                for ( i = 0 ; i < N ; i += n ) {

                        for ( j = 0 ; j < n/2 ; j++ ) {

                                cexp.r = cos( isign*2*PI*j/n);
                                cexp.i = sin( isign*2*PI*j/n);


                                A1 = &A[i+j];
                                A2 = &A[i+j+n/2];

                                temp.r = cexp.r*A2->r - cexp.i*A2->i;
                                temp.i = cexp.r*A2->i + cexp.i*A2->r;

                                A2->r = A1->r - temp.r;
                                A2->i = A1->i - temp.i;

                                A1->r = A1->r + temp.r;
                                A1->i = A1->i + temp.i;

                                }

                        }
                }

}







/* NOTES ON PROGRESSING TO R2DIT.C:
 *
 * The following code fragments show how to progress to the production
 * r2dit.c from the basic unrolled fft code given above.
 * 
 * The objective is to rework the butterfly section to reduce the number
 * of cos and sin evaluations.  These optimizations result in a routine
 * with only small lookup tables and no calls to trig functions.
 * 
 * The first step is to exchange the i and j loops in fft.c to avoid
 * redundant trig evaluations on the innermost loop.
 * 
 * 
 *         for ( n = 2 ; n <= N ; n *= 2 ) {
 * 
 * 
 *                 for ( j = 0 ; j < n/2 ; j++ ) {
 * 
 *                         cexp.r = cos( isign*2*PI*j/n);
 *                         cexp.i = sin( isign*2*PI*j/n);
 * 
 * 
 *                         for ( i = 0 ; i < N ; i += n ) {
 * 
 *                                 A1 = &A[i+j];
 *                                 A2 = &A[i+j+n/2];
 * 
 *                                 temp.r = cexp.r*A2->r - cexp.i*A2->i;
 *                                 temp.i = cexp.r*A2->i + cexp.i*A2->r;
 * 
 *                                 A2->r = A1->r - temp.r;
 *                                 A2->i = A1->i - temp.i;
 * 
 *                                 A1->r = A1->r + temp.r;
 *                                 A1->i = A1->i + temp.i;
 * 
 *                                 }
 * 
 *                         }
 *                 }
 * 
 * 
 * 
 * 
 * 
 * 
 * The next step moves cos and sin outside the j loop, by twiddling their
 * values along with the trig addition formulas.  Twiddling is less work
 * than evaluating cos and sin.
 * 
 * The following fragment lists the twiddled butterfly section.
 * 
 *         COMPLEX twid;
 * 
 * 
 *         for ( n = 2 ; n <= N ; n *= 2 ) {
 * 
 *                 twid.r = cos( isign*2*PI/n);
 *                 twid.i = sin( isign*2*PI/n);
 *             
 *                 cexp.r = 1.0;
 *                 cexp.i = 0.0;
 * 
 *                 for ( j = 0 ; j < n/2 ; j++ ) {
 * 
 * 
 *                         for ( i = 0 ; i < N ; i += n ) {
 * 
 *                                 A1 = &A[i+j];
 *                                 A2 = &A[i+j+n/2];
 * 
 *                                 temp.r = cexp.r*A2->r - cexp.i*A2->i;
 *                                 temp.i = cexp.r*A2->i + cexp.i*A2->r;
 * 
 *                                 A2->r = A1->r - temp.r;
 *                                 A2->i = A1->i - temp.i;
 * 
 *                                 A1->r = A1->r + temp.r;
 *                                 A1->i = A1->i + temp.i;
 * 
 *                                 }
 * 
 * 
 *                         Twiddle cexp ...
 * 
 *                         temp.r = twid.r*cexp.r - twid.i*cexp.i;
 *                         temp.i = twid.i*cexp.r + twid.r*cexp.i;
 * 
 *                         cexp.r = temp.r;
 *                         cexp.i = temp.i;
 * 
 *                         }
 * 
 *                 }
 * 
 * 
 * Production code is usually further optimized by taking the starting
 * twid values from a lookup table, removing the need for cos and sin
 * entirely.  Note that the tables don't have to be very big to support
 * even large FFTs.
 * 
 * The resulting code can be found in r2dit.c ...
 * 
 * 
 * 
 * 
 */
