/* FILE: r2dit.c                Copyright (c), Symmetric Research, 1986 1998
 *
 * Production level radix 2, decimation in time, in place FFT.
 *
 *     F[k] = sum( A[n] * cexp( isign*2*PI*n*k/N))    n,k = 0,1,...,N-1
 *
 *
 *
 * NOTES:
 *
 * This is a production level, radix 2, decimation in time, FFT.  See
 * recurse.c and loop.c for development leading to this routine.
 *
 * No trig functions are required for this function.  Small lookup tables
 * are used instead.  Trim their size to the minimum dimension required.
 *
 * Further optimizations to this code are often taking bit reverse pairs
 * from a lookup table, and unrolling the butterfly loops to be inline
 * code.  Generally, these steps result in the code size growing quickly
 * and yield only small performance improvements.
 *
 * For integer applications, often the best optimization step to pursue
 * next is converting the function to use fixed point arithmetic.  This
 * is particularly true if the computer is emulating floating point.
 *
 *
 *
 *
 *
 *
 */

#include <math.h>
#include "fft.h"







/* FFT twiddle tables:
 *
 * These are the complex twiddle values for the cexp term in the radix 2
 * FFT butterflys.
 *
 *
 *
 */

COMPLEX

        FORWARD_TWIDS[] = {

                /* cos(2*PI/n), sin( 2*PI/n) */

                -1.00000000e+00,  0.00000000e-00,    /* n = 000002 */
                 0.00000000e-00,  1.00000000e+00,    /* n = 000004 */
                 7.07106781e-01,  7.07106781e-01,    /* n = 000008 */
                 9.23879533e-01,  3.82683432e-01,    /* n = 000016 */
                 9.80785280e-01,  1.95090322e-01,    /* n = 000032 */
                 9.95184727e-01,  9.80171403e-02,    /* n = 000064 */
                 9.98795456e-01,  4.90676743e-02,    /* n = 000128 */
                 9.99698819e-01,  2.45412285e-02,    /* n = 000256 */
                 9.99924702e-01,  1.22715383e-02,    /* n = 000512 */
                 9.99981175e-01,  6.13588465e-03,    /* n = 001024 */
                 9.99995294e-01,  3.06795676e-03,    /* n = 002048 */
                 9.99998823e-01,  1.53398019e-03,    /* n = 004096 */
                 9.99999706e-01,  7.66990319e-04,    /* n = 008192 */
                 9.99999926e-01,  3.83495188e-04,    /* n = 016384 */
                 9.99999982e-01,  1.91747597e-04,    /* n = 032768 */
                 9.99999995e-01,  9.58737991e-05,    /* n = 065536 */
                 9.99999999e-01,  4.79368996e-05,    /* n = 131072 */

                },


        INVERSE_TWIDS[] = {

                /* cos(2*PI/n), -sin( 2*PI/n) */

                -1.00000000e+00,  0.00000000e-00,    /* n = 000002 */
                 0.00000000e-00, -1.00000000e+00,    /* n = 000004 */
                 7.07106781e-01, -7.07106781e-01,    /* n = 000008 */
                 9.23879533e-01, -3.82683432e-01,    /* n = 000016 */
                 9.80785280e-01, -1.95090322e-01,    /* n = 000032 */
                 9.95184727e-01, -9.80171403e-02,    /* n = 000064 */
                 9.98795456e-01, -4.90676743e-02,    /* n = 000128 */
                 9.99698819e-01, -2.45412285e-02,    /* n = 000256 */
                 9.99924702e-01, -1.22715383e-02,    /* n = 000512 */
                 9.99981175e-01, -6.13588465e-03,    /* n = 001024 */
                 9.99995294e-01, -3.06795676e-03,    /* n = 002048 */
                 9.99998823e-01, -1.53398019e-03,    /* n = 004096 */
                 9.99999706e-01, -7.66990319e-04,    /* n = 008192 */
                 9.99999926e-01, -3.83495188e-04,    /* n = 016384 */
                 9.99999982e-01, -1.91747597e-04,    /* n = 032768 */
                 9.99999995e-01, -9.58737991e-05,    /* n = 065536 */
                 9.99999999e-01, -4.79368996e-05,    /* n = 131072 */

                };









/* FFT function:
 *
 * If isign is greater than zero a forward transform is done, otherwise an
 * inverse.
 *
 *
 *
 *
 */

void FFT( COMPLEX *A, int N, int isign) {

        int n, i, j, mask; COMPLEX temp, *A1, *A2, *t, cexp;
        



        /* Scramble the input data into bit reversed index order. */

        for ( i = 0 , j = 0 ; i < N ; i++ ) {


                /* Swap A[i] and A[j] if not done yet. */

                if ( j > i ) {

                        A1 = &A[i];
                        A2 = &A[j];

                        temp.r = A1->r;
                        temp.i = A1->i;
                        
                        A1->r = A2->r;
                        A1->i = A2->i;
                        
                        A2->r = temp.r;
                        A2->i = temp.i;

                        }


                /* Reverse inc j. */

                for ( mask = N/2 ; mask&j ; mask >>= 1 )

                        j &= ~mask;

                j |= mask;

                }





        /* Compute the FFT butterflies. */

        t = ( isign > 0 ? FORWARD_TWIDS : INVERSE_TWIDS );


        for ( n = 2 ; n <= N ; n *= 2 ) {

                cexp.r = 1.0;
                cexp.i = 0.0;

                for ( j = 0 ; j < n/2 ; j++ ) {


                        for ( i = 0 ; i < N ; i += n ) {

                                A1 = &A[i+j];
                                A2 = &A[i+j+n/2];

                                temp.r = cexp.r*A2->r - cexp.i*A2->i;
                                temp.i = cexp.r*A2->i + cexp.i*A2->r;

                                A2->r = A1->r - temp.r;
                                A2->i = A1->i - temp.i;

                                A1->r = A1->r + temp.r;
                                A1->i = A1->i + temp.i;

                                }


                        /* Twiddle cexp. */

                        temp.r = t->r*cexp.r - t->i*cexp.i;
                        temp.i = t->i*cexp.r + t->r*cexp.i;

                        cexp.r = temp.r;
                        cexp.i = temp.i;

                        }



                /* Move to the next twiddle factor. */

                t++;

                }

}
