/* FILE: recurse.c              Copyright (c), Symmetric Research, 1986 1998
 *
 * Theoretical recursive radix 2, decimation in time, in place FFT.
 *
 *     F[k] = sum( A[n] * cexp( isign*2*PI*n*k/N))    n,k = 0,1,...,N-1
 *
 *
 *
 * NOTES:
 *
 * This code is a recursive implementation of the basic theoretical
 * radix 2 decimation in time FFT formula.  
 * 
 * The function is very memory and cpu intensive.  Do not use it for
 * production applications.  See loop.c and r2dit.c for more typical
 * non recursive production code.
 *
 * Besides being a useful step in FFT development, this routine can also
 * be used for code generation for fully unlooped inline FFTs.
 *
 * The recursion chews up stack.  You may need to compile with more than
 * the usual defaults.  Also consider changing the locals to static.
 *
 *
 *
 *
 */

#include <math.h>
#include "fft.h"            /* << for the define of COMPLEX */




#define MAXFFT  512

COMPLEX even[MAXFFT/2], odd[MAXFFT/2];




void FFT( COMPLEX *A, int N, int isign) {

        int i, k; COMPLEX temp, *A1, *A2, cexp;

                /* these really chew up stack on the recurse ... */

        


        /* Check the input dimension. */

        if ( N > MAXFFT || N == 1 )

                return;




        /* Scramble the input array into even elements followed by odd. */

        for ( i = 0 ; i < N/2 ; i++ ) {

                /* Copy out the even and odd elements. */

                even[i].r = A[2*i].r;
                even[i].i = A[2*i].i;

                 odd[i].r = A[2*i+1].r;
                 odd[i].i = A[2*i+1].i;

                }


        for ( i = 0 ; i < N/2 ; i++ ) {


                /* Repack A with all the even elements first. */

                A[i].r = even[i].r;
                A[i].i = even[i].i;


                /* Followed by all the odd elements. */

                A[i+N/2].r = odd[i].r;
                A[i+N/2].i = odd[i].i;

                }





        /* Recurse, computing FFTs of even and odd elements respectively. */

        if ( N > 2 )

                { FFT( A, N/2, isign); FFT( A+N/2, N/2, isign); }




        /* Stitch them together with a butterfly operation. */

        for ( k = 0 ; k < N/2 ; k++ ) {

                A1 = &A[k];
                A2 = &A[k+N/2];

                cexp.r = cos( (isign*2*PI*k)/N);
                cexp.i = sin( (isign*2*PI*k)/N);

                temp.r = cexp.r*A2->r - cexp.i*A2->i;
                temp.i = cexp.r*A2->i + cexp.i*A2->r;

                A2->r = A1->r - temp.r;
                A2->i = A1->i - temp.i;

                A1->r = A1->r + temp.r;
                A1->i = A1->i + temp.i;

                }
}



/* NOTES ON PROGRESSING TO LOOP.C:
 *
 * When unrolling this code to remove the recursion, the even/odd
 * shuffle steps before the recursive dimension N/2 FFTs become a
 * scramble into bit reversed order.  To see this, think about the low
 * bit of the index measuring whether the entry is even or odd.  Since
 * all the even entries need to go to the first half of the vector, and
 * the odd to the high half, the low bit in effect becomes the high bit.
 * The process is repeated for all the bits in the index as you recurse
 * on down.
 *
 *
 *
 */
